# Copyright 2017 - Nokia Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from glare.common import exception as exc
from glare.tests import sample_artifact
from glare.tests.unit import base

import random


class TestArtifactList(base.BaseTestArtifactAPI):

    def test_list_simple_fields(self):
        # Create a bunch of artifacts for list testing
        values = [
            {'name': 'art1', 'version': '0.0.1', 'string_required': 'str1',
             'int1': 5, 'float1': 5.0, 'bool1': 'yes'},
            {'name': 'art1', 'version': '1-beta', 'string_required': 'str2',
             'int1': 6, 'float1': 6.0, 'bool1': 'yes'},
            {'name': 'art1', 'version': '1', 'string_required': 'str1',
             'int1': 5, 'float1': 5.0, 'bool1': 'no', 'description': 'ggg'},
            {'name': 'art1', 'version': '2-rc1', 'string_required': 'str22',
             'int1': 7, 'float1': 7.0, 'bool1': 'yes'},
            {'name': 'art1', 'version': '10', 'string_required': 'str222',
             'int1': 5, 'float1': 5.0, 'bool1': 'yes'},
            {'name': 'art2', 'version': '1', 'string_required': 'str1',
             'int1': 8, 'float1': 8.0, 'bool1': 'no'},
            {'name': 'art3', 'version': '1', 'string_required': 'str1',
             'int1': -5, 'float1': -5.0, 'bool1': 'yes'},
        ]
        arts = [self.controller.create(self.req, 'sample_artifact', val)
                for val in values]

        # Activate 3rd and 4th artifacts
        changes = [{'op': 'replace', 'path': '/status', 'value': 'active'}]
        arts[3] = self.update_with_values(changes, art_id=arts[3]['id'])
        arts[4] = self.update_with_values(changes, art_id=arts[4]['id'])

        # Publish 4th artifact
        changes = [{'op': 'replace', 'path': '/visibility', 'value': 'public'}]
        self.req = self.get_fake_request(user=self.users['admin'])
        arts[4] = self.update_with_values(changes, art_id=arts[4]['id'])
        self.req = self.get_fake_request(user=self.users['user1'])

        # Do tests basic tests
        # input format for filters is a list of tuples:
        # (filter_name, filter_value)

        # List all artifacts
        res = self.controller.list(self.req, 'sample_artifact')
        self.assertEqual(7, len(res['artifacts']))
        self.assertEqual('sample_artifact', res['type_name'])

        # List all artifacts as an anonymous. Only public artifacts are visible
        anon_req = self.get_fake_request(user=self.users['anonymous'])
        res = self.controller.list(anon_req, 'sample_artifact')
        self.assertEqual(1, len(res['artifacts']))
        self.assertIn(arts[4], res['artifacts'])

        # Filter by name
        filters = [('name', 'art1')]
        res = self.controller.list(self.req, 'sample_artifact', filters)
        self.assertEqual(5, len(res['artifacts']))

        filters = [('name', 'in:art2,art3')]
        res = self.controller.list(self.req, 'sample_artifact', filters)
        self.assertEqual(2, len(res['artifacts']))
        for i in (5, 6):
            self.assertIn(arts[i], res['artifacts'])

        # Filter by string_required
        filters = [('string_required', 'str1')]
        res = self.controller.list(self.req, 'sample_artifact', filters)
        self.assertEqual(4, len(res['artifacts']))
        for i in (0, 2, 5, 6):
            self.assertIn(arts[i], res['artifacts'])

        # Filter by int1
        filters = [('int1', '5')]
        res = self.controller.list(self.req, 'sample_artifact', filters)
        self.assertEqual(3, len(res['artifacts']))
        for i in (0, 2, 4):
            self.assertIn(arts[i], res['artifacts'])

        filters = [('int1', 'in:5,6')]
        res = self.controller.list(self.req, 'sample_artifact', filters)
        self.assertEqual(4, len(res['artifacts']))
        for i in (0, 1, 2, 4):
            self.assertIn(arts[i], res['artifacts'])

        # Filter by float1
        filters = [('float1', '5.0')]
        res = self.controller.list(self.req, 'sample_artifact', filters)
        self.assertEqual(3, len(res['artifacts']))
        for i in (0, 2, 4):
            self.assertIn(arts[i], res['artifacts'])

        # Filter by bool1
        filters = [('bool1', 'yes')]
        res = self.controller.list(self.req, 'sample_artifact', filters)
        self.assertEqual(5, len(res['artifacts']))
        for i in (0, 1, 3, 4, 6):
            self.assertIn(arts[i], res['artifacts'])

        # Filter by id
        filters = [('id', arts[0]['id'])]
        res = self.controller.list(self.req, 'sample_artifact', filters)
        self.assertEqual(1, len(res['artifacts']))
        self.assertIn(arts[0], res['artifacts'])

        # Filter by status
        filters = [('status', 'active')]
        res = self.controller.list(self.req, 'sample_artifact', filters)
        self.assertEqual(2, len(res['artifacts']))
        for i in (3, 4):
            self.assertIn(arts[i], res['artifacts'])

        # Filter by visibility
        filters = [('visibility', 'public')]
        res = self.controller.list(self.req, 'sample_artifact', filters)
        self.assertEqual(1, len(res['artifacts']))
        self.assertIn(arts[4], res['artifacts'])

        # Filter by owner
        filters = [('owner', arts[0]['owner'])]
        res = self.controller.list(self.req, 'sample_artifact', filters)
        self.assertEqual(7, len(res['artifacts']))
        for i in range(6):
            self.assertIn(arts[i], res['artifacts'])

        # Filter by description leads to BadRequest
        filters = [('description', 'ggg')]
        self.assertRaises(exc.BadRequest, self.controller.list,
                          self.req, 'sample_artifact', filters)

        # Filter by created_at with eq operator leads to BadRequest
        filters = [('created_at', arts[4]['created_at'])]
        self.assertRaises(exc.BadRequest, self.controller.list,
                          self.req, 'sample_artifact', filters)

        # Filter by updated_at with eq operator leads to BadRequest
        filters = [('updated_at', arts[4]['updated_at'])]
        self.assertRaises(exc.BadRequest, self.controller.list,
                          self.req, 'sample_artifact', filters)

        # Filter by activated_at with eq operator leads to BadRequest
        filters = [('activated_at', arts[4]['activated_at'])]
        self.assertRaises(exc.BadRequest, self.controller.list,
                          self.req, 'sample_artifact', filters)

        # Filter by any blob leads to BadRequest
        filters = [('blob', 'something')]
        self.assertRaises(exc.BadRequest, self.controller.list,
                          self.req, 'sample_artifact', filters)

        # Filter by nonexistent field leads to BadRequest
        filters = [('NONEXISTENT', 'something')]
        self.assertRaises(exc.BadRequest, self.controller.list,
                          self.req, 'sample_artifact', filters)

    def test_list_marker_and_limit(self):
        # Create artifacts
        art_list = [
            self.controller.create(
                self.req, 'sample_artifact',
                {'name': 'name%s' % i,
                 'version': '%d.0' % i,
                 'tags': ['tag%s' % i],
                 'int1': 1024 + i,
                 'float1': 123.456,
                 'str1': 'bugaga',
                 'bool1': True})
            for i in range(5)]

        # sort with 'next_marker'
        sort = [('int1', 'asc'), ('name', 'desc')]
        result = self.controller.list(self.req, 'sample_artifact', filters=(),
                                      limit=1, sort=sort)
        self.assertEqual([art_list[0]], result['artifacts'])
        marker = result['next_marker']
        result = self.controller.list(self.req, 'sample_artifact', filters=(),
                                      marker=marker, limit=1, sort=sort)
        self.assertEqual([art_list[1]], result['artifacts'])

        # sort by custom marker
        sort = [('int1', 'asc')]
        marker = art_list[1]['id']
        result = self.controller.list(self.req, 'sample_artifact', filters=(),
                                      marker=marker, sort=sort)
        self.assertEqual(art_list[2:], result['artifacts'])

        sort = [('int1', 'desc')]
        result = self.controller.list(self.req, 'sample_artifact', filters=(),
                                      marker=marker, sort=sort)
        self.assertEqual(art_list[:1], result['artifacts'])

        sort = [('float1', 'asc'), ('name', 'desc')]
        result = self.controller.list(self.req, 'sample_artifact', filters=(),
                                      marker=marker, sort=sort)
        self.assertEqual([art_list[0]], result['artifacts'])

        # paginate by name in desc order with limit 2
        sort = [('name', 'desc')]
        result = self.controller.list(self.req, 'sample_artifact', filters=(),
                                      limit=2, sort=sort)
        self.assertEqual(art_list[4:2:-1], result['artifacts'])

        marker = result['next_marker']
        result = self.controller.list(self.req, 'sample_artifact', filters=(),
                                      marker=marker, limit=2, sort=sort)
        self.assertEqual(art_list[2:0:-1], result['artifacts'])

        marker = result['next_marker']
        result = self.controller.list(self.req, 'sample_artifact', filters=(),
                                      marker=marker, limit=2, sort=sort)
        self.assertEqual([art_list[0]], result['artifacts'])

        # paginate by version in desc order with limit 2
        sort = [('version', 'desc')]
        result = self.controller.list(self.req, 'sample_artifact', filters=(),
                                      limit=2, sort=sort)
        self.assertEqual(art_list[4:2:-1], result['artifacts'])

        marker = result['next_marker']
        result = self.controller.list(self.req, 'sample_artifact', filters=(),
                                      marker=marker, limit=2, sort=sort)
        self.assertEqual(art_list[2:0:-1], result['artifacts'])

        marker = result['next_marker']
        result = self.controller.list(self.req, 'sample_artifact', filters=(),
                                      marker=marker, limit=2, sort=sort)
        self.assertEqual([art_list[0]], result['artifacts'])

    def test_list_version(self):
        values = [
            {'name': 'art1', 'version': '0.0.1'},
            {'name': 'art1', 'version': '1-beta'},
            {'name': 'art1', 'version': '1'},
            {'name': 'art1', 'version': '10-rc1'},
            {'name': 'art1', 'version': '10'},
            {'name': 'art2', 'version': '1'},
            {'name': 'art3', 'version': '1'},
        ]

        arts = [self.controller.create(self.req, 'sample_artifact', val)
                for val in values]

        # List all artifacts
        res = self.controller.list(self.req, 'sample_artifact', [])
        self.assertEqual(7, len(res['artifacts']))
        self.assertEqual('sample_artifact', res['type_name'])

        # Get latest artifacts
        res = self.controller.list(self.req, 'sample_artifact', [],
                                   latest=True)
        self.assertEqual(3, len(res['artifacts']))
        for i in (4, 5, 6):
            self.assertIn(arts[i], res['artifacts'])

        # Various version filters
        filters = [('version', '1')]
        res = self.controller.list(self.req, 'sample_artifact', filters)
        self.assertEqual(3, len(res['artifacts']))
        for i in (2, 5, 6):
            self.assertIn(arts[i], res['artifacts'])

        filters = [('version', '1'), ('name', 'art1')]
        res = self.controller.list(self.req, 'sample_artifact', filters)
        self.assertEqual(1, len(res['artifacts']))
        self.assertIn(arts[2], res['artifacts'])

        filters = [('version', 'gt:1')]
        res = self.controller.list(self.req, 'sample_artifact', filters)
        self.assertEqual(2, len(res['artifacts']))
        for i in (3, 4):
            self.assertIn(arts[i], res['artifacts'])

        filters = [('version', 'gte:1')]
        res = self.controller.list(self.req, 'sample_artifact', filters)
        self.assertEqual(5, len(res['artifacts']))
        for i in (2, 3, 4, 5, 6):
            self.assertIn(arts[i], res['artifacts'])

        filters = [('version', 'lte:1')]
        res = self.controller.list(self.req, 'sample_artifact', filters)
        self.assertEqual(5, len(res['artifacts']))
        for i in (0, 1, 2, 5, 6):
            self.assertIn(arts[i], res['artifacts'])

        filters = [('version', 'gt:1-beta'), ('version', 'lt:10')]
        res = self.controller.list(self.req, 'sample_artifact', filters)
        self.assertEqual(4, len(res['artifacts']))
        for i in (2, 3, 5, 6):
            self.assertIn(arts[i], res['artifacts'])

        filters = [('version', 'in:0.0.1,10-rc1')]
        res = self.controller.list(self.req, 'sample_artifact', filters)
        self.assertEqual(2, len(res['artifacts']))
        for i in (0, 3):
            self.assertIn(arts[i], res['artifacts'])

        # Filter by invalid version
        filters = [('version', 'INVALID_VERSION')]
        self. assertRaises(exc.BadRequest, self.controller.list,
                           self.req, 'sample_artifact', filters)

        # Filter by invalid operator
        filters = [('version', 'INVALID_op:1')]
        self. assertRaises(exc.BadRequest, self.controller.list,
                           self.req, 'sample_artifact', filters)

    def test_list_compound_fields(self):
        # Create a bunch of artifacts for list testing
        values = [
            {'name': 'art1',
             'dict_of_str': {'a': 'aa', 'b': 'bb'},
             'dict_of_int': {'one': 1, 'two': 2},
             'list_of_str': ['aa', 'bb'],
             'list_of_int': [1, 2]},
            {'name': 'art2',
             'dict_of_str': {'b': 'bb', 'c': 'cc'},
             'dict_of_int': {'two': 2, 'three': 3},
             'list_of_str': ['bb', 'cc'],
             'list_of_int': [2, 3]},
            {'name': 'art3',
             'dict_of_str': {'a': 'aa', 'c': 'cc'},
             'dict_of_int': {'one': 1, 'three': 3},
             'list_of_str': ['aa', 'cc'],
             'list_of_int': [1, 3]},
            {'name': 'art4',
             'dict_of_str': {'a': 'bb'},
             'dict_of_int': {'one': 2},
             'list_of_str': ['aa'],
             'list_of_int': [1]},
            {'name': 'art5',
             'dict_of_str': {'b': 'bb'},
             'dict_of_int': {'two': 2},
             'list_of_str': ['bb'],
             'list_of_int': [2]},
            {'name': 'art6',
             'dict_of_str': {},
             'dict_of_int': {},
             'list_of_str': [],
             'list_of_int': []},
        ]
        arts = [self.controller.create(self.req, 'sample_artifact', val)
                for val in values]

        # List all artifacts
        res = self.controller.list(self.req, 'sample_artifact', [])
        self.assertEqual(6, len(res['artifacts']))
        self.assertEqual('sample_artifact', res['type_name'])

        # Return artifacts that contain key 'a' in 'dict_of_str'
        filters = [('dict_of_str', 'eq:a')]
        res = self.controller.list(self.req, 'sample_artifact', filters)
        self.assertEqual(3, len(res['artifacts']))
        for i in (0, 2, 3):
            self.assertIn(arts[i], res['artifacts'])

        # Return artifacts that contain key 'a' or 'c' in 'dict_of_str'
        filters = [('dict_of_str', 'in:a,c')]
        res = self.controller.list(self.req, 'sample_artifact', filters)
        self.assertEqual(4, len(res['artifacts']))
        for i in (0, 1, 2, 3):
            self.assertIn(arts[i], res['artifacts'])

        # Filter with invalid operator leads to BadRequest
        filters = [('dict_of_str', 'invalid:a')]
        self.assertRaises(exc.BadRequest, self.controller.list,
                          self.req, 'sample_artifact', filters)

        # Return artifacts that contain key one in 'dict_of_int'
        filters = [('dict_of_int', 'eq:one')]
        res = self.controller.list(self.req, 'sample_artifact', filters)
        self.assertEqual(3, len(res['artifacts']))
        for i in (0, 2, 3):
            self.assertIn(arts[i], res['artifacts'])

        # Return artifacts that contain key one or three in 'dict_of_int'
        filters = [('dict_of_int', 'in:one,three')]
        res = self.controller.list(self.req, 'sample_artifact', filters)
        self.assertEqual(4, len(res['artifacts']))
        for i in (0, 1, 2, 3):
            self.assertIn(arts[i], res['artifacts'])

        # Filter by dicts values
        # Return artifacts that contain value 'bb' in 'dict_of_str[b]'
        filters = [('dict_of_str.b', 'eq:bb')]
        res = self.controller.list(self.req, 'sample_artifact', filters)
        self.assertEqual(3, len(res['artifacts']))
        for i in (0, 1, 4):
            self.assertIn(arts[i], res['artifacts'])

        # Return artifacts that contain values 'aa' or 'bb' in 'dict_of_str[a]'
        filters = [('dict_of_str.a', 'in:aa,bb')]
        res = self.controller.list(self.req, 'sample_artifact', filters)
        self.assertEqual(3, len(res['artifacts']))
        for i in (0, 2, 3):
            self.assertIn(arts[i], res['artifacts'])

        # Filter with invalid operator leads to BadRequest
        filters = [('dict_of_str.a', 'invalid:aa')]
        self.assertRaises(exc.BadRequest, self.controller.list,
                          self.req, 'sample_artifact', filters)

        # Return artifacts that contain value '2' in 'dict_of_int[two]'
        filters = [('dict_of_int.two', 'eq:2')]
        res = self.controller.list(self.req, 'sample_artifact', filters)
        self.assertEqual(3, len(res['artifacts']))
        for i in (0, 1, 4):
            self.assertIn(arts[i], res['artifacts'])

        # Return artifacts that contain values '1' or '2' in 'dict_of_int[one]'
        filters = [('dict_of_int.one', 'in:1,2')]
        res = self.controller.list(self.req, 'sample_artifact', filters)
        self.assertEqual(3, len(res['artifacts']))
        for i in (0, 2, 3):
            self.assertIn(arts[i], res['artifacts'])

        # Filter with invalid operator leads to BadRequest
        filters = [('dict_of_int.one', 'invalid:1')]
        self.assertRaises(exc.BadRequest, self.controller.list,
                          self.req, 'sample_artifact', filters)

        # Filter by nonexistent dict leads to BadRequest
        filters = [('NOTEXIST.one', 'eq:1')]
        self.assertRaises(exc.BadRequest, self.controller.list,
                          self.req, 'sample_artifact', filters)

        # Test with TypeError
        filters = [('dict_of_int.1', 'lala')]
        self.assertRaises(exc.BadRequest, self.controller.list,
                          self.req, 'sample_artifact', filters)

        # Return artifacts that contain key 'aa' in 'list_of_str'
        filters = [('list_of_str', 'eq:aa')]
        res = self.controller.list(self.req, 'sample_artifact', filters)
        self.assertEqual(3, len(res['artifacts']))
        for i in (0, 2, 3):
            self.assertIn(arts[i], res['artifacts'])

        # Return artifacts that contain key 'aa' or 'cc' in 'list_of_str'
        filters = [('list_of_str', 'in:aa,cc')]
        res = self.controller.list(self.req, 'sample_artifact', filters)
        self.assertEqual(4, len(res['artifacts']))
        for i in (0, 1, 2, 3):
            self.assertIn(arts[i], res['artifacts'])

        # Filter with invalid operator leads to BadRequest
        filters = [('list_of_str', 'invalid:aa')]
        self.assertRaises(exc.BadRequest, self.controller.list,
                          self.req, 'sample_artifact', filters)

        # Return artifacts that contain key 1 in 'list_of_int'
        filters = [('list_of_int', 'eq:1')]
        res = self.controller.list(self.req, 'sample_artifact', filters)
        self.assertEqual(3, len(res['artifacts']))
        for i in (0, 2, 3):
            self.assertIn(arts[i], res['artifacts'])

        # Return artifacts that contain key 1 or three in 'list_of_int'
        filters = [('list_of_int', 'in:1,3')]
        res = self.controller.list(self.req, 'sample_artifact', filters)
        self.assertEqual(4, len(res['artifacts']))
        for i in (0, 1, 2, 3):
            self.assertIn(arts[i], res['artifacts'])

    def test_filter_by_tags(self):
        values = [
            {'name': 'name1', 'tags': ['tag1', 'tag2']},
            {'name': 'name2', 'tags': ['tag1', 'tag3']},
            {'name': 'name3', 'tags': ['tag1']},
            {'name': 'name4', 'tags': ['tag2']},
            {'name': 'name5', 'tags': ['tag4']},
            {'name': 'name6', 'tags': ['tag4', 'tag5']},
        ]
        arts = [self.controller.create(self.req, 'sample_artifact', val)
                for val in values]

        filters = [('tags', 'tag1')]
        res = self.controller.list(self.req, 'sample_artifact', filters)
        self.assertEqual(3, len(res['artifacts']))
        for i in (0, 1, 2):
            self.assertIn(arts[i], res['artifacts'])

        filters = [('tags', 'tag1,tag2')]
        res = self.controller.list(self.req, 'sample_artifact', filters)
        self.assertEqual(1, len(res['artifacts']))
        self.assertIn(arts[0], res['artifacts'])

        filters = [('tags', 'NOT_A_TAG')]
        res = self.controller.list(self.req, 'sample_artifact', filters)
        self.assertEqual(0, len(res['artifacts']))

        filters = [('tags-any', 'tag1')]
        res = self.controller.list(self.req, 'sample_artifact', filters)
        self.assertEqual(3, len(res['artifacts']))
        for i in (0, 1, 2):
            self.assertIn(arts[i], res['artifacts'])

        filters = [('tags-any', 'tag1,NOT_A_TAG')]
        res = self.controller.list(self.req, 'sample_artifact', filters)
        self.assertEqual(3, len(res['artifacts']))
        for i in (0, 1, 2):
            self.assertIn(arts[i], res['artifacts'])

        filters = [('tags-any', 'tag2,tag5')]
        res = self.controller.list(self.req, 'sample_artifact', filters)
        self.assertEqual(3, len(res['artifacts']))
        for i in (0, 3, 5):
            self.assertIn(arts[i], res['artifacts'])

        # Filtering by tags with operators leads to BadRequest
        for f in ('tags', 'tags-any'):
            filters = [(f, 'eq:tag1')]
            self.assertRaises(
                exc.BadRequest, self.controller.list,
                self.req, 'sample_artifact', filters)

    def test_list_and_sort_fields(self):
        amount = 7
        # Create a bunch of artifacts for list sorting tests
        names = random.sample(["art%d" % i for i in range(amount)], amount)
        floats = random.sample([0.01 * i for i in range(amount)], amount)
        ints = random.sample([1 * i for i in range(amount)], amount)
        strings = random.sample(["str%d" % i for i in range(amount)], amount)
        versions = random.sample(["0.%d" % i for i in range(amount)], amount)
        for i in range(amount):
            val = {'name': names[i], 'float1': floats[i], 'int1': ints[i],
                   'str1': strings[i], 'version': versions[i]}
            self.controller.create(self.req, 'sample_artifact', val)

        fields = ['name', 'id', 'visibility', 'version', 'float1', 'int1',
                  'str1']

        for sort_name in fields:
            for sort_dir in ['asc', 'desc']:
                arts = self.controller.list(
                    self.req, 'sample_artifact', [],
                    sort=[(sort_name, sort_dir)])['artifacts']
                self.assertEqual(amount, len(arts))
                sorted_arts = sorted(arts, key=lambda x: x[sort_name],
                                     reverse=sort_dir == 'desc')
                self.assertEqual(sorted_arts, arts)

    def test_list_and_sort_negative(self):
        # sort by non-existent field
        self.assertRaises(exc.BadRequest, self.controller.list,
                          self.req, 'sample_artifact',
                          [], sort=[("NONEXISTENT", "desc")])

        # sort by wrong direction
        self.assertRaises(exc.BadRequest, self.controller.list,
                          self.req, 'sample_artifact',
                          [], sort=[("name", "WRONG_DIR")])

        # For performance sake sorting by more than one custom field
        # is forbidden. Nevertheless, sorting by several basic field are
        # absolutely fine.
        # List of basic fields is located in glare/db/sqlalchemy/api.py as
        # BASE_ARTIFACT_PROPERTIES tuple.
        sort = [("int1", "desc"), ("float1", "desc")]
        self.assertRaises(exc.BadRequest, self.controller.list,
                          self.req, 'sample_artifact',
                          [], sort=sort)

        # sort with non-sortable fields
        for name, field in sample_artifact.SampleArtifact.fields.items():
            for sort_dir in ['asc', 'desc']:
                if not field.sortable:
                    self.assertRaises(
                        exc.BadRequest, self.controller.list,
                        self.req, 'sample_artifact',
                        [], sort=[(name, sort_dir)])



# Copyright (c) 2016 Mirantis, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import functools

from glare.api.v1 import api_version_request as api_version
from glare.common import exception as exc
from glare.i18n import _


class VersionedMethod(object):

    def __init__(self, name, start_version, end_version, func):
        """Versioning information for a single method.

        :param name: Name of the method
        :param start_version: Minimum acceptable version
        :param end_version: Maximum acceptable_version
        :param func: Method to call
        """
        # NOTE(kairat): minimums and maximums are inclusive
        self.name = name
        self.start_version = start_version
        self.end_version = end_version
        self.func = func

    def __str__(self):
        return ("Version Method %s: min: %s, max: %s"
                % (self.name, self.start_version, self.end_version))


class VersionedResource(object):
    """Versioned mixin that provides ability to define versioned methods and
    return appropriate methods based on user request.
    """

    # prefix for all versioned methods in class
    VER_METHODS_ATTR_PREFIX = 'versioned_methods_'

    @staticmethod
    def check_for_versions_intersection(func_list):
        """Determines whether function list contains version intervals
        intersections or not. General algorithm:
        https://en.wikipedia.org/wiki/Intersection_algorithm

        :param func_list: list of VersionedMethod objects
        :return: boolean
        """
        pairs = []
        counter = 0
        for f in func_list:
            pairs.append((f.start_version, 1, f))
            pairs.append((f.end_version, -1, f))

        def compare(x):
            return x[0]

        pairs.sort(key=compare)
        for p in pairs:
            counter += p[1]
            if counter > 1:
                return True
        return False

    @classmethod
    def supported_versions(cls, min_ver, max_ver=None):
        """Decorator for versioning api methods.

        Add the decorator to any method which takes a request object
        as the first parameter and belongs to a class which inherits from
        wsgi.Controller. The implementation inspired by Nova.

        :param min_ver: string representing minimum version
        :param max_ver: optional string representing maximum version
        """

        def decorator(f):
            obj_min_ver = api_version.APIVersionRequest(min_ver)
            if max_ver:
                obj_max_ver = api_version.APIVersionRequest(max_ver)
            else:
                obj_max_ver = api_version.APIVersionRequest.max_version()

            # Add to list of versioned methods registered
            func_name = f.__name__
            new_func = VersionedMethod(func_name, obj_min_ver, obj_max_ver, f)

            versioned_attr = cls.VER_METHODS_ATTR_PREFIX + cls.__name__
            func_dict = getattr(cls, versioned_attr, {})
            if not func_dict:
                setattr(cls, versioned_attr, func_dict)

            func_list = func_dict.get(func_name, [])
            if not func_list:
                func_dict[func_name] = func_list
                func_list.append(new_func)

            # Ensure the list is sorted by minimum version (reversed)
            # so later when we work through the list in order we find
            # the method which has the latest version which supports
            # the version requested.
            is_intersect = cls.check_for_versions_intersection(
                func_list)

            if is_intersect:
                raise exc.ApiVersionsIntersect(
                    name=new_func.name,
                    min_ver=new_func.start_version,
                    max_ver=new_func.end_version,
                )

            func_list.sort(key=lambda vf: vf.start_version, reverse=True)

            return f

        return decorator

    def __getattribute__(self, key):
        def version_select(*args, **kwargs):
            """Look for the method which matches the name supplied and version
            constraints and calls it with the supplied arguments.

            :returns: Returns the result of the method called
            :raises: VersionNotFoundForAPIMethod if there is no method which
             matches the name and version constraints
            """
            # versioning is used in 3 classes: request deserializer and
            # controller have request as first argument
            # response serializer has response as first argument
            # we must respect all three cases
            if hasattr(args[0], 'api_version_request'):
                ver = args[0].api_version_request
            elif hasattr(args[0], 'request'):
                ver = args[0].request.api_version_request
            else:
                raise exc.VersionNotFoundForAPIMethod(
                    message=_("Api version not found in the request."))

            func_list = self.versioned_methods[key]
            for func in func_list:
                if ver.matches(func.start_version, func.end_version):
                    # Update the version_select wrapper function so
                    # other decorator attributes like wsgi.response
                    # are still respected.
                    functools.update_wrapper(version_select, func.func)
                    return func.func(self, *args, **kwargs)

            # No version match
            raise exc.VersionNotFoundForAPIMethod(version=ver)

        class_obj = object.__getattribute__(self, '__class__')
        prefix = object.__getattribute__(self, 'VER_METHODS_ATTR_PREFIX')
        attr_name = prefix + object.__getattribute__(class_obj, '__name__')
        try:
            if key in object.__getattribute__(self, attr_name):
                return version_select
        except AttributeError:
            # No versioning on this class
            pass

        return object.__getattribute__(self, key)



# Copyright 2016 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import collections
import importlib
import pkgutil
import sys

from oslo_config import cfg
from oslo_config import types as conf_types
from oslo_log import log as logging
from oslo_versionedobjects import base as vo_base

from glare.common import exception
from glare.i18n import _
from glare.objects import base

CONF = cfg.CONF

LOG = logging.getLogger(__name__)

registry_options = [
    cfg.ListOpt('enabled_artifact_types',
                default=['heat_templates', 'heat_environments',
                         'murano_packages', 'tosca_templates', 'images'],
                item_type=conf_types.String(),
                help=_("List of enabled artifact types that will be "
                       "available to user")),
    cfg.ListOpt('custom_artifact_types_modules', default=[],
                item_type=conf_types.String(),
                help=_("List of custom user modules with artifact types that "
                       "will be uploaded by Glare dynamically during service "
                       "startup."))
]
CONF.register_opts(registry_options)


def import_submodules(module):
    """Import all submodules of a module.

    :param module: Package name
    :return: list of imported modules
    """
    package = sys.modules[module]
    return [
        importlib.import_module(module + '.' + name)
        for loader, name, is_pkg in pkgutil.walk_packages(package.__path__)]


def import_modules_list(modules):
    custom_module_list = []
    for module_name in modules:
        try:
            custom_module_list.append(importlib.import_module(module_name))
        except Exception as e:
            LOG.exception(e)
            LOG.error("Cannot import custom artifact type from module "
                      "%(module_name)%s. Error: %(error)s",
                      {'module_name': module_name, 'error': str(e)})
    return custom_module_list


def get_subclasses(module, base_class):
    subclasses = []
    for name in dir(module):
        obj = getattr(module, name)
        try:
            if issubclass(obj, base_class) and obj != base_class:
                subclasses.append(obj)
        except TypeError:
            pass
    return subclasses


class ArtifactRegistry(vo_base.VersionedObjectRegistry):
    """Artifact Registry is responsible for registration of artifacts and
    returning appropriate artifact types based on artifact type name.
    """

    @classmethod
    def register_all_artifacts(cls):
        """Register all artifacts in Glare."""
        # get all submodules in glare.objects
        # please note that we registering trusted modules first
        # and applying custom modules after that to allow custom modules
        # to specify custom logic inside
        modules = (import_submodules('glare.objects') +
                   import_modules_list(
                       CONF.custom_artifact_types_modules))
        # get all versioned object classes in module
        supported_types = []
        for module in modules:
            supported_types.extend(get_subclasses(module, base.BaseArtifact))

        for type_name in set(CONF.enabled_artifact_types + ['all']):
            for af_type in supported_types:
                if type_name == af_type.get_type_name():
                    if af_type != 'all':
                        CONF.register_opts(
                            af_type.list_artifact_type_opts(),
                            group='artifact_type:' + type_name)
                    cls.register(af_type)
                    break
            else:
                raise exception.TypeNotFound(name=type_name)

    @classmethod
    def get_artifact_type(cls, type_name):
        """Return artifact type based on artifact type name.

        :param type_name: name of artifact type
        :return: artifact class
        """
        for name, af_type in cls.obj_classes().items():
            if af_type[0].get_type_name() == type_name:
                return af_type[0]
        raise exception.TypeNotFound(name=type_name)

    @classmethod
    def reset_registry(cls):
        """Resets all registered artifact type classes."""
        cls._registry._obj_classes = collections.defaultdict(list)



# Copyright 2017 - Nokia Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from glare.tests.functional import base


class TestVisibility(base.TestArtifact):

    """Test Glare artifact visibility for various users."""

    def test_visibility_basic(self):

        self.set_user('user1')
        art1 = self.create_artifact(data={'name': 'art1', 'version': 1.0})
        url = '/sample_artifact/%s' % art1['id']

        # Artifact is visible by its owner
        self.get(url=url)

        # Owner can modify the artifact
        patch = [{"op": "replace", "path": "/description", "value": "dd"}]
        self.patch(url=url, data=patch)

        # Artifact is not visible by another user
        self.set_user('user2')
        self.get(url=url, status=404)

        # Artifact is visible by admin
        self.set_user('admin')
        self.get(url=url)

        # Admin can update the artifact
        patch = [{"op": "replace", "path": "/string_required", "value": "gg"}]
        self.patch(url=url, data=patch)

        # Activate and publish the artifact
        self.patch(url=url, data=self.make_active)
        self.patch(url=url, data=self.make_public)

        # All users can see public artifact

        self.set_user('user1')
        self.get(url=url)

        # Default policy 'update_public' forbids the owner to update public
        # artifacts
        patch = [{"op": "replace", "path": "/description", "value": "bb"}]
        self.patch(url=url, data=patch, status=403)

        self.set_user('admin')
        self.get(url=url)

        # Admin can always update public artifacts
        patch = [{"op": "replace", "path": "/description", "value": "ss"}]
        self.patch(url=url, data=patch)

        self.set_user('user2')
        self.get(url=url)

        # Regular user cannot update public artifact
        patch = [{"op": "replace", "path": "/description", "value": "aa"}]
        self.patch(url=url, data=patch, status=403)

    def test_visibility_name_version(self):
        self.set_user('user1')
        self.create_artifact(data={'name': 'my_art', 'version': 1.0})

        # User can't create another artifact with the same name/version
        self.create_artifact(data={'name': 'my_art', 'version': 1.0},
                             status=409)

        art2 = self.create_artifact(data={'name': 'your_art', 'version': 2.0})
        url = '/sample_artifact/%s' % art2['id']

        # User can't change name and version if such artifact already exists
        patch = [
            {"op": "replace", "path": "/name", "value": "my_art"},
            {"op": "replace", "path": "/version", "value": 1.0}
        ]
        self.patch(url=url, data=patch, status=409)

        # Another user can create an artifact with the same name/version
        self.set_user("user2")
        art3 = self.create_artifact(data={'name': 'my_art', 'version': 1.0})

        # Now admin sees 2 artifacts with the same name/version
        self.set_user("admin")
        url = '/sample_artifact?name=my_art&version=1'
        self.assertEqual(2, len(self.get(url=url)['sample_artifact']))

        # Admin can activate and publish artifact art3
        url = '/sample_artifact/%s' % art3['id']
        patch = [{"op": "replace", "path": "/string_required", "value": "gg"}]
        self.patch(url=url, data=patch)
        self.patch(url=url, data=self.make_active)
        self.patch(url=url, data=self.make_public)

        # After that user1 sees 2 artifacts with the same name/version as well
        self.set_user("user1")
        url = '/sample_artifact?name=my_art&version=1'
        self.assertEqual(2, len(self.get(url=url)['sample_artifact']))

        # User2 still sees only his public artifact
        self.set_user("user2")
        url = '/sample_artifact?name=my_art&version=1'
        self.assertEqual(1, len(self.get(url=url)['sample_artifact']))

        # Admin is able to create a private artifact with the same name/version
        self.set_user("admin")
        art4 = self.create_artifact(data={'name': 'my_art', 'version': 1.0})

        # And he sees 3 artifacts
        url = '/sample_artifact?name=my_art&version=1'
        self.assertEqual(3, len(self.get(url=url)['sample_artifact']))

        # But he can't publish his artifact, because this name/version already
        # exists in public scope
        url = '/sample_artifact/%s' % art4['id']
        patch = [{"op": "replace", "path": "/string_required", "value": "gg"}]
        self.patch(url=url, data=patch)
        self.patch(url=url, data=self.make_active)
        self.patch(url=url, data=self.make_public, status=409)

        # Admin publishes artifact art2
        url = '/sample_artifact/%s' % art2['id']
        patch = [{"op": "replace", "path": "/string_required", "value": "gg"}]
        self.patch(url=url, data=patch)
        self.patch(url=url, data=self.make_active)
        self.patch(url=url, data=self.make_public)

        # User2 can create his own private artifact with the same name/version
        self.set_user("user2")
        self.create_artifact(data={'name': 'your_art', 'version': 2.0})

    def test_visibility_artifact_types(self):
        self.set_user('user1')
        self.create_artifact(data={'name': 'my_art', 'version': 1.0},
                             type_name='images')
        self.create_artifact(data={'name': 'my_art', 'version': 1.0},
                             type_name='heat_templates')
        self.create_artifact(data={'name': 'my_art', 'version': 1.0},
                             type_name='heat_environments')

    def test_visibility_all(self):
        self.set_user('user1')
        art1 = self.create_artifact(data={'name': 'my_art', 'version': 1.0},
                                    type_name='images')
        art2 = self.create_artifact(data={'name': 'my_art', 'version': 1.0},
                                    type_name='heat_templates')
        # User 1 sees his 2 artifacts
        url = '/all?name=my_art&version=1'
        self.assertEqual(2, len(self.get(url=url)['all']))

        self.set_user('user2')
        self.create_artifact(data={'name': 'my_art', 'version': 1.0},
                             type_name='images')
        self.create_artifact(data={'name': 'my_art', 'version': 1.0},
                             type_name='heat_templates')
        # User 2 sees his 2 artifacts
        url = '/all?name=my_art&version=1'
        self.assertEqual(2, len(self.get(url=url)['all']))

        # Admin sees 4 artifacts from both users
        self.set_user("admin")
        self.assertEqual(4, len(self.get(url=url)['all']))

        # After publishing art1 and art2 user 2 can see 4 artifacts as well
        url = '/sample_artifact/%s' % art1['id']
        patch = [{"op": "replace", "path": "/string_required", "value": "gg"}]
        self.patch(url=url, data=patch)
        self.patch(url=url, data=self.make_active)
        self.patch(url=url, data=self.make_public)

        url = '/sample_artifact/%s' % art2['id']
        patch = [{"op": "replace", "path": "/string_required", "value": "gg"}]
        self.patch(url=url, data=patch)
        self.patch(url=url, data=self.make_active)
        self.patch(url=url, data=self.make_public)

        self.set_user("user2")
        url = '/all?name=my_art&version=1'
        self.assertEqual(4, len(self.get(url=url)['all']))
